/*! \mainpage Template Overlap documentation
 *
 * \section intro_sec Introduction
 *
 *
 *  TemplateOverlap is a  C++ package that provides a basic implementation
of the Template Overlap Method for jet substructure, as well as several jet analysis tools.
The program allows the user to define a template model by specifying a model for a signal.
The templates are calculated with a script provided in the package. The program then allows you to seach numerically for
the template that gives the best match to a given jet. Additional commands enable you to further
study jet substructure, by e.g. computing jet shapes for those jets.  TemplateOverlap combines
the flexibility of energy flow methods, with the speed of a numerical package.
 *
 * Many of the facilities of TemplateOverlap can be accessed through the two following classes:
 *   - TemplateOverlap::Event:  The basic class for holding the 4-momentum of the jet constituents or templates.
 *   - MatchingMethod:  The main class for template matching analyses
 *
 *
 * \section install_sec Installation
 *

 * The TemplateOverlap package can be set up and run as follows.
 *
 * Download the code and then unpack it:
 * \verbatim
  wget http://www.hepforge.org/archive/tom/TemplateOverlap-0.3.0.tar.gz
  tar zxvf TemplateOverlap-0.3.0.tar.gz
   \endverbatim

 * Move to the source folder
 *   \verbatim

 cd TemplateOverlap-0.3.0/
     \endverbatim

 and then compile and link the example  example.cc that comes with the tarball
  *   \verbatim

  g++ -Wall -O3 example3.cc -o example3
   \endverbatim

 */

#include "matching.hh"
#include "TemplateTagger.hh"

#include "fastjet/FunctionOfPseudoJet.hh"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <time.h> // For timing
#include <iomanip> 
#include <stdio.h> // For pipe



using namespace std;

using namespace TemplateOverlap;

unsigned int WhichPtBin (double PtJet, double PtMin, double PtBinWidth, int nptbins)
{
if (PtJet < PtMin) return 0;
int i =  int(floor  ((PtJet - PtMin)/PtBinWidth));
if (i > nptbins) i = nptbins;
if (i<0) return 0;
return i;
}



///  Helper function for printing jets
void PrintJets(const vector <fastjet::PseudoJet>& jets) {
   printf("%5s %15s %15s %15s %15s %15s\n","jet #", "rapidity", "phi", "pt","m","e"); // label the columns
   for (unsigned int i = 0; i < jets.size(); i++) {
      printf("%5u %15.8f %15.8f %15.8f %15.8f %15.8f\n",i, jets[i].rap(),jets[i].phi(),jets[i].perp(),jets[i].m(),jets[i].e());
   }
}


//----------------------------------------------------------------------
// forward declaration for printing out info about a jet
//----------------------------------------------------------------------
ostream & operator<<(ostream &, const fastjet::PseudoJet &);


//----------------------------------------------------------------------
// class for reading in a jet
//----------------------------------------------------------------------

bool ReadEvent (ifstream& is,  
		vector<fastjet::PseudoJet> & input_particles, 
		vector<fastjet::PseudoJet> & hard_partons)
{

  string line;
  double pup1, pup2, pup3, pup4;
  int nhep, i, isthep,idhep,evtnumber;
  
  input_particles.resize(0);
  hard_partons.resize(0);
  
  if (!getline(is,line)) return false;	
  istringstream getpro(line);
  getpro  >> evtnumber >> nhep;

  for (int j=0; j<nhep;++j){
    if (!getline(is,line)) return false;	
    istringstream getall(line);
    getall  >> i >> isthep >>idhep>> pup1 >> pup2 >> pup3 >> pup4;
	
    if ( isthep == 21 )
    {
      fastjet::PseudoJet pJet = fastjet::PseudoJet(pup1,pup2,pup3,pup4);
      pJet.set_user_index(idhep);  
      hard_partons.push_back(pJet);
    } 

    if (fabs(idhep) == 12 || fabs(idhep) == 14 || fabs(idhep) == 16) continue;
    
     if ( isthep == 1 )
    {
      fastjet::PseudoJet pJet = fastjet::PseudoJet(pup1,pup2,pup3,pup4);
      pJet.set_user_index(idhep);  
      input_particles.push_back(pJet);
    } 
  } 

return evtnumber;

}

void splitPartons(const std::vector<fastjet::PseudoJet> & partons, std::vector<fastjet::PseudoJet> & bquarks, std::vector<fastjet::PseudoJet> & leptons)
{
  bquarks.resize(0);
  leptons.resize(0);
  for (unsigned int i=0;i<partons.size();++i)
  {
    int id = abs(partons[i].user_index());
    if (id==5) bquarks.push_back(partons[i]);
    if (id>=11 && id<=14) leptons.push_back(partons[i]);
  }
  
}

//---------------------------------------------------------
/////////////////Add pileup
//---------------------------------------------------------

void addPileup(std::vector<fastjet::PseudoJet> & particles, std::ifstream & pileupFile, const int nPileup)
{
  std::vector <fastjet::PseudoJet> leptons, pileups;
  for (int iPileup = 0; iPileup < nPileup; ++iPileup) 
  {
    if(!ReadEvent(pileupFile,pileups, leptons)) 
    {
      pileupFile.clear() ;
      pileupFile.seekg(0, ios::beg) ;
      ReadEvent(pileupFile,pileups, leptons);
      
    }
    for (unsigned int i = 0; i < pileups.size(); ++i)
      particles.push_back(pileups[i]);
  }
}

int main(int argc, char* argv[]) {
  
  if (argc !=5) {
    cout << "Usage:  "<< argv[0] << " <input.dat> <template2.dat> <template3.dat> <pileup.dat>" << endl;
    return 1;
  }
   /// Set Run Parameters 
  double subConeRadius = 0.30;
  double subConeRadius3 = 0.12;
  double coneRadius = 1.2;
  double sigma = 0.333 ;
  double ovcut = 0.; /// For TemplateTagger
  int nPileup = 0; /// For pileup overlay
 
 /// Create best matched template containers
  std::vector<fastjet::PseudoJet> maxTempl;

 /// Create event containers
  std::vector<fastjet::PseudoJet> input_particles, partons;
  std::vector<fastjet::PseudoJet> myHiggs, myW;


 /// Where to read the events from
  ifstream file(argv[1]);
  ifstream pileupFile(argv[4]);
  
  /// Set up the template tagger
  fastjet::SharedPtr<fastjet::Noverlap> cone(new fastjet::Noverlap(subConeRadius, sigma, argv[2], FIXED, CONE));
  fastjet::TemplateTagger tagger(cone.get(), fastjet::SelectorMassRange(110,250), subConeRadius, ovcut);
  
  /// Set up the template tagger
  fastjet::SharedPtr<fastjet::Noverlap> cone3(new fastjet::Noverlap(subConeRadius3, sigma, argv[3], VARIABLE, CONE));
  fastjet::TemplateTagger filter(cone3.get(), fastjet::SelectorMassRange(110,250), subConeRadius3, ovcut);
  
  /// read in data in format i isthep idhep px py pz E 

  while(ReadEvent(file, input_particles, partons)){
   clock_t start = clock(); // start clock

   /// Jet definitions 
  fastjet::JetDefinition         *jetDef = NULL;
  jetDef = new fastjet::JetDefinition(fastjet::antikt_algorithm, coneRadius,
                                      fastjet::E_scheme, fastjet::Best);
  /// Overlay pileup
  if (nPileup > 0) addPileup(input_particles, pileupFile, nPileup);
  
   /// Find antikt jets
  vector <fastjet::PseudoJet> fatjets;
  fastjet::ClusterSequence clustSeq(input_particles, *jetDef);    
  fatjets = sorted_by_pt(clustSeq.inclusive_jets(25.));
  
  cout << "Ran: " << jetDef->description() << endl << endl;
  cout << "Hardest jet: " << fatjets[0] << endl << endl;

  
  /// Now tag the highest-pT jet using template tagger
  fastjet::PseudoJet tagged = tagger(fatjets[0]);

  /// Show me what you got
  if (tagged == 0){
    cout << "No substructure found" << endl;
  }  
  else{
    
  /// Filter the jet with 3-parton templates
  
  fastjet::PseudoJet filtered = filter(fatjets[0]);
  if (filtered == 0){
    cout << "No 3-body substructure found" << endl;
  }  

  
  fastjet::PseudoJet parent1 = tagged.pieces()[0];
  fastjet::PseudoJet parent2 = tagged.pieces()[1];
  cout << "Found suitable pair of subjets: " << std::endl;
  cout << " " << parent1 << endl;
  cout << " " << parent2 << endl;
  cout << "Total = " << endl;
  cout << " " << tagged << endl;
  cout << "(Ov2 = " << tagged.structure_of<fastjet::TemplateTagger>().ov() <<")"
 
  << endl << endl;
  cout << " The best-matched templates are: "<< std::endl;     
  PrintJets(tagged.structure_of<fastjet::TemplateTagger>().maxTempl());
  
  std::vector<fastjet::PseudoJet> my2bTemplate = tagged.structure_of<fastjet::TemplateTagger>().maxTempl();
  
  
  /// Create display file

  std::ofstream fout("sample_event_display.dat",std::ifstream::out );   
  std::ofstream fout2("sample_event_out.dat",std::ifstream::out );  
  std::ofstream fout3("sample_event_parton.dat",std::ifstream::out );  
    
  splitPartons(partons, myHiggs, myW);
  
   
  double Ov3 = (filtered == 0) ? 0.0 : filtered.structure_of<fastjet::TemplateTagger>().ov();
  
  std::stringstream ss; 
  ss << "(Ov2 = "<<tagged.structure_of<fastjet::TemplateTagger>().ov() <<" ";
  if (Ov3 >0.0) ss << ", Ov3 = "<< Ov3  ;
  ss << ")" ;

 
  if (filtered != 0){
  std::vector<fastjet::PseudoJet> my3bTemplate = filtered.structure_of<fastjet::TemplateTagger>().maxTempl();
  
  double r1 = my3bTemplate[0].user_index()/100000.;
  double r2 = my3bTemplate[1].user_index()/100000.;
  double r3 = my3bTemplate[2].user_index()/100000.;

 
  fastjet::planarflow pf(r1,r2,r3);
  fastjet::planarflow pf2(0,0,0);
  
  
  fastjet::PseudoJet axis3b = fastjet::join(my3bTemplate);
  
  cout << "Jet Pf = " << pf2(fatjets[0]) << std::endl;
  cout << " Template Pf = " << pf(axis3b) << std::endl;

  
  }
 
 
  std::cout << "Time Elapsed: " << ((double)clock() - start) / CLOCKS_PER_SEC << std::endl;

  std::string mystr;
  std::cout << "Please enter (y) to continue or (q) to abort the program:  ";
  std::getline (cin, mystr);
  if (mystr == "q") return 0;
  
  }
  }

  return 0;
}


//----------------------------------------------------------------------
// does the actual work for printing out a jet
//----------------------------------------------------------------------
ostream & operator<<(ostream & ostr, const fastjet::PseudoJet & jet) {
  ostr << "pt, y, phi =" 
       << " " << setw(10) << jet.perp() 
       << " " << setw(6) <<  jet.rap()  
       << " " << setw(6) <<  jet.phi()  
       << ", mass = " << setw(10) << jet.m()
       << ", btag = " << jet.user_index();
  return ostr;
	
	
}
